/*
 [ GPL ]
 
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com> 

 http://kldp.net
 http://minzkn.pe.ky
*/

#ifndef DEF_SOURCE_mixer_c 
#define DEF_SOURCE_mixer_c "mixer.c"

#ifndef DEF_STANDALONE 
#define DEF_STANDALONE (1)
#endif

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#if DEF_STANDALONE == (0)
#include "mixer.h"
#else
#include <stdio.h>
#include <string.h>

#define DEF_MIXER_GET  (0)
#define DEF_MIXER_SET  (1)
#define DEF_MIXER_KEEP (-1)
#endif

static const char *__MZ_MixerDeviceName__(const char *s_DeviceName);
static int __MZ_MixerOpen__(const char *s_DeviceName);
static void __MZ_MixerClose__(int s_Handle);
static int __MZ_MixerGet__(int s_Handle, int s_ChannelIndex, int *s_Level);
static int __MZ_MixerSet__(int s_Handle, int s_ChannelIndex, int *s_Level);

int MZ_GetMixer(const char *s_DeviceName, int s_Channel, int *s_Left, int *s_Right, int *s_Mute);
int MZ_GetVolume(const char *s_DeviceName, int s_Channel, int *s_Left, int *s_Right);
int MZ_GetMuteStatus(const char *s_DeviceName, int s_Channel);
int MZ_GetMute(const char *s_DeviceName, int s_Channel);
int MZ_SetMixer(const char *s_DeviceName, int s_Channel, int s_Left, int s_Right, int s_Mute);
int MZ_SetVolume(const char *s_DeviceName, int s_Channel, int s_Left, int s_Right);
int MZ_SetMuteStatus(const char *s_DeviceName, int s_Channel, int s_Mute);
int MZ_SetMute(const char *s_DeviceName, int s_Channel, int s_Mute);
int MZ_SetMixerMask(const char *s_DeviceName, int s_ChannelMask, int s_Left, int s_Right, int s_Mute);
int MZ_SetVolumeMask(const char *s_DeviceName, int s_ChannelMask, int s_Left, int s_Right);
int MZ_SetMuteStatusMask(const char *s_DeviceName, int s_ChannelMask, int s_Mute);
int MZ_SetMuteMask(const char *s_DeviceName, int s_ChannelMask, int s_Mute);

const int c_SoundDevices = SOUND_MIXER_NRDEVICES;
const char *c_MixerName[] = SOUND_DEVICE_NAMES;
char *g_DefaultMixerDeviceName = "/dev/mixer";

/* [ Static function ] ==================================================================== */
static const char *__MZ_MixerDeviceName__(const char *s_DeviceName)
{
 return(s_DeviceName ? s_DeviceName : g_DefaultMixerDeviceName);
}

static int __MZ_MixerOpen__(const char *s_DeviceName)
{
 return(open(__MZ_MixerDeviceName__(s_DeviceName), O_RDONLY));
}

static void __MZ_MixerClose__(int s_Handle)
{
 if(s_Handle >= 0)close(s_Handle);
}

static int __MZ_MixerGet__(int s_Handle, int s_ChannelIndex, int *s_Level)
{
 int s_Return = (-1);
 if(ioctl(s_Handle, MIXER_READ(s_ChannelIndex), s_Level, sizeof(int)) == 0)s_Return = 1;
 return(s_Return);
}

static int __MZ_MixerSet__(int s_Handle, int s_ChannelIndex, int *s_Level)
{
 int s_Return = (-1);
 if(ioctl(s_Handle, MIXER_WRITE(s_ChannelIndex), s_Level, sizeof(int)) == 0)s_Return = 1;
 return(s_Return);
}

/* [ Public function ] ==================================================================== */
int MZ_GetMixer(const char *s_DeviceName, int s_Channel, int *s_Left, int *s_Right, int *s_Mute)
{
 int s_Return, s_Handle, s_Level;
 s_Handle = __MZ_MixerOpen__(s_DeviceName);
 if(s_Handle >= 0)
 {
  s_Return = __MZ_MixerGet__(s_Handle, s_Channel, &s_Level);
  if(s_Return > 0)
  {
   if(s_Left )*(s_Left ) = ((s_Level >> 8) & 0x7f) % (100 + 1);
   if(s_Right)*(s_Right) = ((s_Level >> 0) & 0x7f) % (100 + 1);
   if(s_Mute )*(s_Mute ) = (s_Level & 0x8000) ? 1 : 0;
  }
  __MZ_MixerClose__(s_Handle);
 }
 else s_Return = (-1);
 return(s_Return);
}

int MZ_GetVolume(const char *s_DeviceName, int s_Channel, int *s_Left, int *s_Right)
{
 return(MZ_GetMixer(s_DeviceName, s_Channel, s_Left, s_Right, (int *)0));
}

int MZ_GetMuteStatus(const char *s_DeviceName, int s_Channel)
{
 int s_Return, s_IsMute;
 s_Return = MZ_GetMixer(s_DeviceName, s_Channel, (int *)0, (int *)0, &s_IsMute);
 if(s_Return > 0)s_Return = s_IsMute;
 return(s_Return);
} 

int MZ_GetMute(const char *s_DeviceName, int s_Channel)
{
 return(MZ_GetMuteStatus(s_DeviceName, s_Channel));
}

int MZ_SetMixer(const char *s_DeviceName, int s_Channel, int s_Left, int s_Right, int s_Mute)
{
 int s_Return, s_Handle, s_Level, s_PreLevel;
 s_Handle = __MZ_MixerOpen__(s_DeviceName);
 if(s_Handle >= 0)
 {
  if(__MZ_MixerGet__(s_Handle, s_Channel, &s_PreLevel) > 0)
  {
   s_Level = 0;
   if(s_Mute > 0)s_Level |= 0x8000;
   else if(s_Mute == 0)s_Level &= 0x7fff;
   else s_Level |= s_PreLevel & 0x8000; /* DEF_MIXER_KEEP */
   if(s_Left  >= 0)s_Level |= s_Left  > 100 ? (100 << 8) : (s_Left << 8);
   else s_Level |= s_PreLevel & 0x7f00; /* DEF_MIXER_KEEP */
   if(s_Right >= 0)s_Level |= s_Right > 100 ?        100 : s_Right;
   else s_Level |= s_PreLevel & 0x007f; /* DEF_MIXER_KEEP */
   s_Return = __MZ_MixerSet__(s_Handle, s_Channel, &s_Level);
  }
  else s_Return = (-1);
  __MZ_MixerClose__(s_Handle);
 }
 else s_Return = (-1);
 return(s_Return);
}

int MZ_SetVolume(const char *s_DeviceName, int s_Channel, int s_Left, int s_Right)
{
 return(MZ_SetMixer(s_DeviceName, s_Channel, s_Left, s_Right, DEF_MIXER_KEEP));
}

int MZ_SetMuteStatus(const char *s_DeviceName, int s_Channel, int s_Mute)
{
 return(MZ_SetMixer(s_DeviceName, s_Channel, DEF_MIXER_KEEP, DEF_MIXER_KEEP, s_Mute));
}

int MZ_SetMute(const char *s_DeviceName, int s_Channel, int s_Mute)
{
 return(MZ_SetMuteStatus(s_DeviceName, s_Channel, s_Mute));
}

int MZ_SetMixerMask(const char *s_DeviceName, int s_ChannelMask, int s_Left, int s_Right, int s_Mute)
{
 int s_Return, s_BitCount, s_BitMask, s_ErrorCount;
 s_Return = s_ErrorCount = 0;
 for(s_BitCount = 0, s_BitMask = 1;s_BitCount < c_SoundDevices && s_BitCount < (sizeof(int) << 3);s_BitCount++, s_BitMask <<= 1)
 {
  if((s_ChannelMask & s_BitMask) == s_BitMask)
  {
   if(MZ_SetMixer(s_DeviceName, s_BitCount, s_Left, s_Right, s_Mute) > 0)s_Return++;
   else s_ErrorCount++;
  }
 }
 if(s_ErrorCount > 0)s_Return = (-1);
 return(s_Return);
}

int MZ_SetVolumeMask(const char *s_DeviceName, int s_ChannelMask, int s_Left, int s_Right)
{
 return(MZ_SetMixerMask(s_DeviceName, s_ChannelMask, s_Left, s_Right, DEF_MIXER_KEEP)); 
}

int MZ_SetMuteStatusMask(const char *s_DeviceName, int s_ChannelMask, int s_Mute)
{
 return(MZ_SetMixerMask(s_DeviceName, s_ChannelMask, DEF_MIXER_KEEP, DEF_MIXER_KEEP, s_Mute)); 
}

int MZ_SetMuteMask(const char *s_DeviceName, int s_ChannelMask, int s_Mute)
{
 return(MZ_SetMuteStatusMask(s_DeviceName, s_ChannelMask, s_Mute));
}

#if DEF_STANDALONE == (1)
static void MixerUsage(void)
{
 fprintf(stdout, "Usage: mixer <channel name> [<Left> [<Right>]]\n"
                "       mixer <channel name> mute <0 | 1>\n"
		"       mixer <channel name>\n"
		"\n");
}

static void ViewMixerChannels(int s_CtrlNo)
{
 int s_Channel, s_Left, s_Right, s_Mute, s_Check, s_CheckCount;
 fprintf(stdout, "==================================\n");
 fprintf(stdout, " %-8s %-10s %-3s %-3s %s\n", "Name", "CtrlNo", "[L]", "[R]", "Mute");
 fprintf(stdout, "----------------------------------\n");
 s_CheckCount = 0;
 if(s_CtrlNo == (-1))
 {
  for(s_Channel = 0;s_Channel < c_SoundDevices;s_Channel++)
  {
   s_Check = MZ_GetMixer((char *)0, s_Channel, &s_Left, &s_Right, &s_Mute);
   if(s_Check > 0)
   {
    s_CheckCount++;
    fprintf(stdout, " %-8s 0x%08x %3d %3d %s\n", c_MixerName[s_Channel], s_Channel, s_Left, s_Right, s_Mute > 0 ? "ON" : "OFF");
   }
  }
 }
 else
 {
  s_Channel = s_CtrlNo;
  s_Check = MZ_GetMixer((char *)0, s_Channel, &s_Left, &s_Right, &s_Mute);
  if(s_Check > 0)
  {
   s_CheckCount++;
   fprintf(stdout, " %-8s 0x%08x %3d %3d %s\n", c_MixerName[s_Channel], s_Channel, s_Left, s_Right, s_Mute > 0 ? "ON" : "OFF");
  }
 }
 if(s_CheckCount == 0)fprintf(stdout, "      [ Not found channel ]\n");
 fprintf(stdout, "==================================\n");
}

static int ChannelNameToCtrlNo(const char *s_ChannelName)
{
 int s_CtrlNo;
 for(s_CtrlNo = 0;s_CtrlNo < c_SoundDevices;s_CtrlNo++)
 {
  if(strcmp(s_ChannelName, c_MixerName[s_CtrlNo]) == 0)return(s_CtrlNo);
 } 
 return(-1);
}

int main(int s_Argc, char *s_Argv[])
{
 fprintf(stdout, "mzmixer release %s - %s %s\n"
                 "Code by JaeHyuk Cho <mailto:minzkn@infoeq.com>, http://kldp.net\n\n"
		 "-----------------------------------------------------------------------\n"
		 "This software comes with ABSOLUTELY NO WARRANTY. This is free software,\n"
		 "and you are welcome to modify and redistribute it under the GPL license\n"
		 "-----------------------------------------------------------------------\n\n",
                 "0.0.9", __DATE__, __TIME__);
 if(s_Argc > 1)
 {
  int s_CtrlNo, s_Left, s_Right, s_Mute;
  s_CtrlNo = ChannelNameToCtrlNo(s_Argv[1]);
  if(s_CtrlNo >= 0)
  {
   if(s_Argc > 2) /* Set */
   {
    if(strcmp(s_Argv[2], "mute") == 0) /* Mute set */
    {
     if(s_Argc > 3)sscanf(s_Argv[3], "%i", &s_Mute);
     else s_Mute = 0; /* Default off */
     if(MZ_SetMute((char *)0, s_CtrlNo, s_Mute) > 0)ViewMixerChannels(s_CtrlNo);
     else fprintf(stdout, "Error: Can not change mute status !!!\n");
    }
    else /* Volume set */
    {
     sscanf(s_Argv[2], "%i", &s_Left);
     if(s_Argc > 3)sscanf(s_Argv[3], "%i", &s_Right);
     else s_Right = s_Left;
     if(MZ_SetVolume((char *)0, s_CtrlNo, s_Left, s_Right) > 0)ViewMixerChannels(s_CtrlNo);
     else fprintf(stdout, "Error: Can not change volume !!!\n");
    }
   }
   else ViewMixerChannels(s_CtrlNo);
  }
  else
  {
   MixerUsage();
   fprintf(stdout, "Invalid argument : Can not found channel name !!!\n");
  } 
 }
 else 
 {
  MixerUsage();
  ViewMixerChannels(-1);
 }
 fprintf(stdout, "\nEnd of mixer.\n");
 return(0);
}
#endif

#endif

/* End of source */
